#include <stdint.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "light.h"
#include "animation.h"
#include "sin.h"

#define CHARGE_PUMP_A_DDR (DDRD)
#define CHARGE_PUMP_A_PIN (PIND)
#define CHARGE_PUMP_A_BIT (0)
#define CHARGE_PUMP_B_DDR (DDRD)
#define CHARGE_PUMP_B_PIN (PIND)
#define CHARGE_PUMP_B_BIT (1)

#define BUTTON_PIN (PINB)
#define BUTTON_BIT (0)
#define BUTTON_PRESS_THRESHOLD (100)
#define BUTTON_HOLD_THRESHOLD (30000)

volatile uint8_t animation_tick_pending = 0;

ISR(TIMER0_OVF_vect) {
	// Toggle the charge pump pins.
	CHARGE_PUMP_A_PIN |= _BV(CHARGE_PUMP_A_BIT);
	CHARGE_PUMP_B_PIN |= _BV(CHARGE_PUMP_B_BIT);
	// Update lights A and B.
	light_timer_update_ab();
}

ISR(TIMER1_OVF_vect) {
	// Update lights C and D.
	light_timer_update_cd();
	// Indicate that we need to animate.
	animation_tick_pending = !0;
}

enum operating_mode {
	OPERATING_MODE_SINGLE,
	OPERATING_MODE_RANDOM,
	OPERATING_MODE_FIXED,
	OPERATING_MODE_COUNT,
};

enum operating_mode mode = OPERATING_MODE_SINGLE;

int main(void) {

	// Configure the charge pump I/O pins as outputs.
	CHARGE_PUMP_A_DDR |= _BV(CHARGE_PUMP_A_BIT);
	CHARGE_PUMP_B_DDR |= _BV(CHARGE_PUMP_B_BIT);

	// 	Invert one half of the charge pump to start with.
	CHARGE_PUMP_A_PIN |= _BV(CHARGE_PUMP_A_BIT);
	
	// Pull the button pin high.
	BUTTON_PIN |= _BV(BUTTON_BIT);

	// Initialise the lights and their animation.
	light_init();
	animation_init();
	
	sei();
	
	static uint16_t button_hold_timer;     // Counts how long the button has been held.
	
	static uint8_t random_change_interval; // Every time this overflows the random change timer is decremented.
	static uint16_t random_change_timer;   // Time until the next random change.
	
	static uint8_t fixed_brightness = 255; // Stores the fixed brightness.
	
	for (;;) {
		
		if (animation_tick_pending) {
			animation_tick_pending = 0;
			
			if (mode == OPERATING_MODE_FIXED) {
				// Set the lights to a fixed brightness.
				for (enum light l = LIGHT_A; l <= LIGHT_D; ++l) {
					light_set_level(l, fixed_brightness);
				}
			} else {
				// Animate the lights.
				animation_tick();
			}
			
			// If we're in random mode, is it time for a change?
			if (random_change_interval++ == 0) {
				if (random_change_timer) {
					if (--random_change_timer == 0) {
						animation_random_type();
						random_change_timer = 500 + (rand() % 2000);
					}
				}
			}
			
			// Are we pressing the button?
			if (!(BUTTON_PIN & _BV(BUTTON_BIT))) {
				if (button_hold_timer != 0xFFFF) {
					++button_hold_timer;
					if (button_hold_timer == BUTTON_HOLD_THRESHOLD) {
						// We're holding the button.
						mode = (mode + 1) % OPERATING_MODE_COUNT;
						// Switch off the random change timer.
						random_change_timer = 0;
					}
					if (button_hold_timer == BUTTON_PRESS_THRESHOLD || button_hold_timer == BUTTON_HOLD_THRESHOLD) {
						// We're pressing the button (or just held it).
						switch (mode) {
							case OPERATING_MODE_SINGLE:
								// Step to the next animation.
								animation_next_type();
								break;
							case OPERATING_MODE_RANDOM:
								// Force a random step.
								random_change_timer = 1;
								random_change_interval = 0;
								break;
							default:
								// Change the fixed brightness.
								if (button_hold_timer != BUTTON_HOLD_THRESHOLD) {
									if (fixed_brightness == 255) {
										fixed_brightness = 0;
									} else {
										fixed_brightness = (fixed_brightness << 1) | 1;
									}
								}
								break;
						}
					}
				}
			} else {
				// No, so reset the counter.
				button_hold_timer = 0;
			}
			
		}
		
	}

}